//////////////////////////////////////////////////////////////////////////////
// Program Name: upnpcdstv.cpp
//                                                                            
// Purpose - uPnp Content Directory Extention for Recorded TV  
//                                                                            
// Created By  : David Blain                    Created On : Jan. 24, 2005
// Modified By :                                Modified On:                  
//                                                                            
//////////////////////////////////////////////////////////////////////////////

#include "upnpcdstv.h"
#include "httprequest.h"
#include <qfileinfo.h>
#include <qregexp.h>
#include <qurl.h>
#include <limits.h>
#include "util.h"

/*
   Recordings                              RecTv
    - All Programs                         RecTv/All
      + <recording 1>                      RecTv/All/item?ChanId=1004&StartTime=2006-04-06T20:00:00
      + <recording 2>
      + <recording 3>
    - By Title                             RecTv/title
      - <title 1>                          RecTv/title/key=Stargate SG-1
        + <recording 1>                    RecTv/title/key=Stargate SG-1/item?ChanId=1004&StartTime=2006-04-06T20:00:00
        + <recording 2>
    - By Genre
    - By Date
    - By Channel
    - By Group
*/


typedef struct
{
    char *title;
    char *column;
    char *sql;
    char *where;

} RootInfo;

static RootInfo g_RootNodes[] = 
{
    {   "All Recordings", 
        "*",
        "SELECT 0 as key, "
          "CONCAT( title, ': ', subtitle) as name, "
          "1 as children "
            "FROM recorded "
            "%1 "
            "ORDER BY starttime DESC",
        "" },

    {   "By Title", 
        "title",
        "SELECT title as id, "
          "title as name, "
          "count( title ) as children "
            "FROM recorded "
            "%1 "
            "GROUP BY title "
            "ORDER BY title", 
        "WHERE title=:KEY" },

    {   "By Genre", 
        "category", 
        "SELECT category as id, "
          "category as name, "
          "count( category ) as children "
            "FROM recorded "
            "%1 "
            "GROUP BY category "
            "ORDER BY category",
        "WHERE category=:KEY" },

    {   "By Date",
        "DATE( starttime )",
        "SELECT DATE( starttime ) as id, "
          "DATE_FORMAT(starttime, '%Y-%m-%d %W') as name, "
          "count( DATE_FORMAT(starttime, '%Y-%m-%d %W') ) as children "
            "FROM recorded "
            "%1 "
            "GROUP BY name "
            "ORDER BY starttime DESC",
        "WHERE DATE( starttime )=:KEY" },

    {   "By Channel", 
        "chanid", 
        "SELECT channel.chanid as id, "
          "CONCAT(channel.channum, ' ', channel.callsign) as name, "
          "count( channum ) as children "
            "FROM channel "
                "INNER JOIN recorded ON channel.chanid = recorded.chanid "
            "%1 "
            "GROUP BY name "
            "ORDER BY channel.chanid",
        "WHERE channel.chanid=:KEY" },


    {   "By Group", 
        "recgroup", 
        "SELECT recgroup as id, "
          "recgroup as name, count( recgroup ) as children "
            "FROM recorded "
            "%1 "
            "GROUP BY recgroup "
            "ORDER BY recgroup",
        "WHERE recgroup=:KEY" }
};

static const short g_nRootNodeLength = sizeof( g_RootNodes ) / sizeof( RootInfo );

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

#define SHARED_RECORDED_SQL "SELECT chanid, starttime, endtime, title, " \
                                   "subtitle, description, category, "   \
                                   "hostname, recgroup, filesize, "      \
                                   "basename, progstart, progend "       \
                             "FROM recorded "


/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

template <class T> inline const T& Min( const T &x, const T &y ) 
{
    return( ( x < y ) ? x : y );
}
 
template <class T> inline const T& Max( const T &x, const T &y ) 
{
    return( ( x > y ) ? x : y );
}
                                                         
/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

UPnpCDSExtensionResults *UPnpCDSTv::Browse( UPnpCDSBrowseRequest *pRequest )
{

    // -=>TODO: Need to add Filter & Sorting Support.
    // -=>TODO: Need to add Sub-Folder/Category Support!!!!!

    if (! pRequest->m_sObjectId.startsWith( m_sExtensionId, true ))
        return( NULL );

    // ----------------------------------------------------------------------
    // Parse out request object's path
    // ----------------------------------------------------------------------

    QStringList idPath = QStringList::split( "/", pRequest->m_sObjectId.section('=',0,0) );

    QString key = pRequest->m_sObjectId.section('=',1);

    if (idPath.count() == 0)
        return( NULL );

    // ----------------------------------------------------------------------
    // Process based on location in hierarchy
    // ----------------------------------------------------------------------

    UPnpCDSExtensionResults *pResults = new UPnpCDSExtensionResults();

    if (pResults != NULL)
    {
        if (key)  
            idPath.last().append(QString("=%1").arg(key)); 

        QString sLast = idPath.last();

        pRequest->m_sParentId = RemoveToken( "/", pRequest->m_sObjectId, 1 );

        if (sLast == m_sExtensionId         ) { return( ProcessRoot   ( pRequest, pResults, idPath )); }
        if (sLast == "0"                    ) { return( ProcessAll    ( pRequest, pResults, idPath )); }
        if (sLast.startsWith( "key" , true )) { return( ProcessKey    ( pRequest, pResults, idPath )); }
        if (sLast.startsWith( "item", true )) { return( ProcessItem   ( pRequest, pResults, idPath )); }

        int nNodeIdx = sLast.toInt();

        if ((nNodeIdx > 0) && (nNodeIdx < g_nRootNodeLength))
            return( ProcessContainer( pRequest, pResults, nNodeIdx, idPath ));

        delete pResults;
    }

    return( NULL );        

}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

QString UPnpCDSTv::RemoveToken( const QString &sToken, const QString &sStr, int num )
{
    QString sResult( "" );
    int     nPos = -1;

    for (int nIdx=0; nIdx < num; nIdx++)
    {
        if ((nPos = sStr.findRev( sToken, nPos )) == -1)
            break;
    }

    if (nPos > 0)
        sResult = sStr.left( nPos );

    return sResult;
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

UPnpCDSExtensionResults *UPnpCDSTv::ProcessRoot( UPnpCDSBrowseRequest    *pRequest, 
                                                 UPnpCDSExtensionResults *pResults, 
                                                 QStringList             &/*idPath*/ )
{
    pResults->m_nTotalMatches   = 0;
    pResults->m_nUpdateID       = 1;

    switch( pRequest->m_eBrowseFlag )
    { 
        case CDS_BrowseMetadata:
        {
            // --------------------------------------------------------------
            // Return Root Object Only
            // --------------------------------------------------------------

            pResults->m_nTotalMatches   = 1;
            pResults->m_nUpdateID       = 1;

            CDSObject *pRoot = CDSObject::CreateContainer( m_sExtensionId, 
                                                           m_sName, 
                                                           "0");

            pRoot->SetChildCount( g_nRootNodeLength );

            pResults->Add( pRoot ); 

            break;
        }

        case CDS_BrowseDirectChildren:
        {
            pResults->m_nUpdateID     = 1;
            pResults->m_nTotalMatches = g_nRootNodeLength;
            
            if ( pRequest->m_nRequestedCount == 0)
                pRequest->m_nRequestedCount = g_nRootNodeLength;

            short nStart = Max( pRequest->m_nStartingIndex, short( 0 ));
            short nEnd   = Min( g_nRootNodeLength, short( nStart + pRequest->m_nRequestedCount));

            if (nStart < g_nRootNodeLength)
            {
                for (short nIdx = nStart; nIdx < nEnd; nIdx++)
                {
                    QString sId = QString( "%1/%2" ).arg( pRequest->m_sObjectId   )
                                                    .arg( nIdx );

                    CDSObject *pItem = CDSObject::CreateContainer( sId,  
                                                                   QObject::tr( g_RootNodes[ nIdx ].title ), 
                                                                   pRequest->m_sParentId );

                    pItem->SetChildCount( GetDistinctCount( g_RootNodes[ nIdx ].column ) );

                    pResults->Add( pItem );
                }
            }
        }

        case CDS_BrowseUnknown:
        default:
            break;
    }

    return( pResults );
}


/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

UPnpCDSExtensionResults *UPnpCDSTv::ProcessAll ( UPnpCDSBrowseRequest    *pRequest,
                                                 UPnpCDSExtensionResults *pResults,
                                                 QStringList             &/*idPath*/ )
{
    // ----------------------------------------------------------------------
    //
    // ----------------------------------------------------------------------
    
    switch( pRequest->m_eBrowseFlag )
    { 
        case CDS_BrowseMetadata:
        {
            // --------------------------------------------------------------
            // Return Container Object Only
            // --------------------------------------------------------------

            pResults->m_nTotalMatches   = 1;
            pResults->m_nUpdateID       = 1;

            CDSObject *pItem = CDSObject::CreateContainer( pRequest->m_sObjectId,  
                                                           QObject::tr( g_RootNodes[ 0 ].title ), 
                                                           m_sExtensionId );
            pItem->SetChildCount( GetDistinctCount( g_RootNodes[ 0 ].column ) );

            pResults->Add( pItem ); 

            break;
        }

        case CDS_BrowseDirectChildren:
        {

            CreateVideoItems( pRequest, pResults, 0, "", false );

            break;
        }

        case CDS_BrowseUnknown:
        default:
            break;

    }
    

    return( pResults );
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

UPnpCDSExtensionResults *UPnpCDSTv::ProcessItem( UPnpCDSBrowseRequest    *pRequest,
                                                 UPnpCDSExtensionResults *pResults, 
                                                 QStringList             &idPath )
{
    pResults->m_nTotalMatches   = 0;
    pResults->m_nUpdateID       = 1;

    // ----------------------------------------------------------------------
    //
    // ----------------------------------------------------------------------
    
    if ( pRequest->m_eBrowseFlag == CDS_BrowseMetadata )
    {
        // --------------------------------------------------------------
        // Return 1 VideoItem
        // --------------------------------------------------------------

        QStringMap  mapParams;
        QString     sParams = idPath.last().section( '?', 1, 1 );
            
        sParams.replace(QRegExp( "&amp;"), "&" ); 

        HTTPRequest::GetParameters( sParams, mapParams );

        int     nChanId    = mapParams[ "ChanId"    ].toInt();
        QString sStartTime = mapParams[ "StartTime" ];

        MSqlQuery query(MSqlQuery::InitCon());

        if (query.isConnected())                                                           
        {
            query.prepare( SHARED_RECORDED_SQL
                           "WHERE chanid=:CHANID and starttime=:STARTTIME " );

            query.bindValue(":CHANID"   , (int)nChanId    );
            query.bindValue(":STARTTIME", sStartTime );
            query.exec();

            if (query.isActive() && query.size() > 0)
            {
                if ( query.next() )
                {
                    pRequest->m_sObjectId = RemoveToken( "/", pRequest->m_sObjectId, 1 );

                    AddVideoItem( pRequest, pResults, false, query );
                    pResults->m_nTotalMatches = 1;
                }
            }
        }
    }

    return( pResults );
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

UPnpCDSExtensionResults *UPnpCDSTv::ProcessKey( UPnpCDSBrowseRequest    *pRequest,
                                                UPnpCDSExtensionResults *pResults, 
                                                QStringList             &idPath )
{
    pResults->m_nTotalMatches   = 0;
    pResults->m_nUpdateID       = 1;

    // ----------------------------------------------------------------------
    //
    // ----------------------------------------------------------------------
    
    QString sKey = idPath.last().section( '=', 1, 1 );
    QUrl::decode( sKey );

    if (sKey.length() > 0)
    {
        int nNodeIdx = idPath[ idPath.count() - 2 ].toInt();

        switch( pRequest->m_eBrowseFlag )
        { 

            case CDS_BrowseMetadata:
            {                                    
                // --------------------------------------------------------------
                // Since Key is not always the title, we need to lookup title.
                // --------------------------------------------------------------

                MSqlQuery query(MSqlQuery::InitCon());

                if (query.isConnected())
                {
                    QString sSQL   = QString( g_RootNodes[ nNodeIdx ].sql )
                                        .arg( g_RootNodes[ nNodeIdx ].where );

                    query.prepare  ( sSQL );
                    query.bindValue( ":KEY", sKey );
                    query.exec();

                    if (query.isActive() && query.size() > 0)
                    {
                        if ( query.next() )
                        {
                            // ----------------------------------------------
                            // Return Container Object Only
                            // ----------------------------------------------

                            pResults->m_nTotalMatches   = 1;
                            pResults->m_nUpdateID       = 1;

                            CDSObject *pItem = CDSObject::CreateContainer( pRequest->m_sObjectId,  
                                                                           query.value(1).toString(), 
                                                                           pRequest->m_sParentId );

                            pItem->SetChildCount( GetDistinctCount( g_RootNodes[ nNodeIdx ].column  ));

                            pResults->Add( pItem ); 
                        }
                    }
                }
                break;
            }

            case CDS_BrowseDirectChildren:
            {
                CreateVideoItems( pRequest, pResults, nNodeIdx, sKey, true );

                break;
            }

            case CDS_BrowseUnknown:
                default:
                break;
        }
    }

    return( pResults );                                                                           
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

UPnpCDSExtensionResults *UPnpCDSTv::ProcessContainer( UPnpCDSBrowseRequest    *pRequest,
                                                      UPnpCDSExtensionResults *pResults,
                                                      int                      nNodeIdx,
                                                      QStringList             &/*idPath*/ )

{
    pResults->m_nUpdateID     = 1;
    pResults->m_nTotalMatches = 0;

    switch( pRequest->m_eBrowseFlag )
    { 
        case CDS_BrowseMetadata:
        {
            // --------------------------------------------------------------
            // Return Container Object Only
            // --------------------------------------------------------------

            pResults->m_nTotalMatches   = 1;
            pResults->m_nUpdateID       = 1;

            CDSObject *pItem = CDSObject::CreateContainer( pRequest->m_sObjectId,  
                                                           QObject::tr( g_RootNodes[ nNodeIdx ].title ), 
                                                           m_sExtensionId );

            pItem->SetChildCount( GetDistinctCount( g_RootNodes[ nNodeIdx ].column ));

            pResults->Add( pItem ); 

            break;
        }

        case CDS_BrowseDirectChildren:
        {
            pResults->m_nTotalMatches = GetDistinctCount( g_RootNodes[ nNodeIdx ].column );
            pResults->m_nUpdateID     = 1;

            if (pRequest->m_nRequestedCount == 0) 
                pRequest->m_nRequestedCount = SHRT_MAX;

            MSqlQuery query(MSqlQuery::InitCon());

            if (query.isConnected())
            {
                // Remove where clause placeholder.

                QString sSQL = g_RootNodes[ nNodeIdx ].sql;
                sSQL.replace( "%1", "" );

                sSQL += QString( " LIMIT %2, %3" )
                           .arg( pRequest->m_nStartingIndex  )
                           .arg( pRequest->m_nRequestedCount );

                query.prepare( sSQL );
                query.exec();

                if (query.isActive() && query.size() > 0)
                {

                    while(query.next())
                    {
                        QString sKey   = query.value(0).toString();
                        QString sTitle = query.value(1).toString();
                        long    nCount = query.value(2).toInt();

                        if (sTitle.length() == 0)
                            sTitle = "(undefined)";

                        QString sId = QString( "%1/%2/key=%3" )
                                         .arg( pRequest->m_sParentId )
                                         .arg( nNodeIdx )
                                         .arg( sKey );

                        CDSObject *pRoot = CDSObject::CreateContainer( sId, sTitle, pRequest->m_sParentId );

                        pRoot->SetChildCount( nCount );

                        pResults->Add( pRoot ); 
                    }
                }
            }

            break;
        }
    }

    return( pResults );
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

int UPnpCDSTv::GetDistinctCount( const QString &sColumn )
{
    int nCount = 0;

    MSqlQuery query(MSqlQuery::InitCon());

    if (query.isConnected())
    {
        // Note: Tried to use Bind, however it would not allow me to use it
        //       for column & table names

        QString sSQL;
        
        if (sColumn == "*")
            sSQL = QString( "SELECT count( %1 ) FROM recorded" ).arg( sColumn );
        else
            sSQL = QString( "SELECT count( DISTINCT %1 ) FROM recorded" ).arg( sColumn );

        query.prepare( sSQL );
        query.exec();

        if (query.size() > 0)
        {
            query.next();

            nCount = query.value(0).toInt();
        }
    }

    return( nCount );
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

int UPnpCDSTv::GetCount( const QString &sColumn, const QString &sKey )
{
    int nCount = 0;

    MSqlQuery query(MSqlQuery::InitCon());

    if (query.isConnected())
    {
        // Note: Tried to use Bind, however it would not allow me to use it
        //       for column & table names

        QString sSQL;
        
        if (sColumn == "*")
            sSQL = "SELECT count( * ) FROM recorded";
        else
            sSQL = QString( "SELECT count( %1 ) FROM recorded WHERE %2=:KEY" )
                      .arg( sColumn )
                      .arg( sColumn );

        query.prepare( sSQL );
        query.bindValue( ":KEY", sKey );
        query.exec();

        if (query.size() > 0)
        {
            query.next();

            nCount = query.value(0).toInt();
        }

    }

    return( nCount );
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

void UPnpCDSTv::CreateVideoItems( UPnpCDSBrowseRequest    *pRequest,
                                  UPnpCDSExtensionResults *pResults,
                                  int                      nNodeIdx,
                                  const QString           &sKey, 
                                  bool                     bAddRef )
{

    pResults->m_nTotalMatches = GetCount( g_RootNodes[ nNodeIdx ].column, sKey );
    pResults->m_nUpdateID     = 1;

    if (pRequest->m_nRequestedCount == 0) 
        pRequest->m_nRequestedCount = SHRT_MAX;

    MSqlQuery query(MSqlQuery::InitCon());

    if (query.isConnected())
    {
        QString sWhere( "" );

        if ( sKey.length() > 0)
        {
           sWhere = QString( "WHERE %1=:KEY " )
                       .arg( g_RootNodes[ nNodeIdx ].column );
        }

        QString sSQL = QString( SHARED_RECORDED_SQL 
                                "%1 "
                                "LIMIT %2, %3" )
                          .arg( sWhere )
                          .arg( pRequest->m_nStartingIndex  )
                          .arg( pRequest->m_nRequestedCount );

        query.prepare  ( sSQL );
        query.bindValue(":KEY", sKey );
        query.exec();

        if (query.isActive() && query.size() > 0)
        {
            while(query.next())
                AddVideoItem( pRequest, pResults, bAddRef, query );
        }
    }
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////

void UPnpCDSTv::AddVideoItem( UPnpCDSBrowseRequest    *pRequest,
                              UPnpCDSExtensionResults *pResults,
                              bool                     bAddRef,
                              MSqlQuery               &query )
{
    int            nChanid      = query.value( 0).toInt();
    QDateTime      dtStartTime  = query.value( 1).toDateTime();
    QDateTime      dtEndTime    = query.value( 2).toDateTime();
    QString        sTitle       = query.value( 3).toString();
    QString        sSubtitle    = query.value( 4).toString();
    QString        sDescription = query.value( 5).toString();
    QString        sCategory    = query.value( 6).toString();
    QString        sHostName    = query.value( 7).toString();
    QString        sRecGroup    = query.value( 8).toString();
    long long      nFileSize    = stringToLongLong( query.value( 9).toString() );
    QString        sBaseName    = query.value(10).toString();
    //VERBOSE(VB_UPNP, QString(" %1 : %2 - %3 - %4").arg(sTitle).arg(sSubtitle).arg(sBaseName).arg(nFileSize));

    QDateTime      dtProgStart  = query.value(11).toDateTime();
    QDateTime      dtProgEnd    = query.value(12).toDateTime();

    // ----------------------------------------------------------------------
    // Cache Host ip Address & Port
    // ----------------------------------------------------------------------

    if (!m_mapBackendIp.contains( sHostName ))
        m_mapBackendIp[ sHostName ] = gContext->GetSettingOnHost( "BackendServerIp", sHostName);

    if (!m_mapBackendPort.contains( sHostName ))
        m_mapBackendPort[ sHostName ] = gContext->GetSettingOnHost("BackendStatusPort", sHostName);

    // ----------------------------------------------------------------------
    // Build Support Strings
    // ----------------------------------------------------------------------

    QString sName      = sTitle + ": " + sSubtitle;

    QString sURIBase   = QString( "http://%1:%2/" )
                            .arg( m_mapBackendIp  [ sHostName ] ) 
                            .arg( m_mapBackendPort[ sHostName ] );

    QString sURIParams = QString( "?ChanId=%1&amp;StartTime=%2" )
                            .arg( nChanid )
                            .arg( dtStartTime.toString(Qt::ISODate));

    QString sId        = QString( "%1/item%2")
                            .arg( pRequest->m_sObjectId )
                            .arg( sURIParams );

    CDSObject *pItem   = CDSObject::CreateVideoItem( sId, 
                                                     sName, 
                                                     pRequest->m_sParentId );
    pItem->m_bRestricted  = false;
    pItem->m_bSearchable  = true;
    pItem->m_sWriteStatus = "WRITABLE";

    if ( bAddRef )
    {
        QString sRefId = QString( "%1/0/item%2")
                            .arg( m_sExtensionId )
                            .arg( sURIParams     );

        pItem->SetPropValue( "refID", sRefId );
    }

    pItem->SetPropValue( "genre"          , sCategory    );
    pItem->SetPropValue( "longDescription", sDescription );
    pItem->SetPropValue( "description"    , sSubtitle    );

    //pItem->SetPropValue( "producer"       , );
    //pItem->SetPropValue( "rating"         , );
    //pItem->SetPropValue( "actor"          , );
    //pItem->SetPropValue( "director"       , );
    //pItem->SetPropValue( "publisher"      , );
    //pItem->SetPropValue( "language"       , );
    //pItem->SetPropValue( "relation"       , );
    //pItem->SetPropValue( "region"         , );

    pResults->Add( pItem );

    // ----------------------------------------------------------------------
    // Add Video Resource Element based on File extension (HTTP)
    // ----------------------------------------------------------------------
    
    QFileInfo fInfo( sBaseName );

    QString sMimeType = HTTPRequest::GetMimeType( fInfo.extension( FALSE ));
    QString sProtocol = QString( "http-get:*:%1:*" ).arg( sMimeType  );
    QString sURI      = QString( "%1getRecording%2").arg( sURIBase   )
                                                    .arg( sURIParams ); 

    Resource *pRes = pItem->AddResource( sProtocol, sURI );

    uint uiStart = dtProgStart.toTime_t();
    uint uiEnd   = dtProgEnd.toTime_t();
    uint uiDur   = uiEnd - uiStart;

    QString sDur = QString( "%1:%2:%3" )
                    .arg( (uiDur / 3600) % 24, 2 )
                    .arg( (uiDur / 60) % 60  , 2 )
                    .arg(  uiDur % 60        , 2 );

    pRes->AddAttribute( "duration"  , sDur      );
    pRes->AddAttribute( "size"      , longLongToString( nFileSize) );


    // ----------------------------------------------------------------------
    // Add Video Resource Element based on File extension (mythtv)
    // ----------------------------------------------------------------------

    sProtocol = QString( "myth:*:%1:*"     ).arg( sMimeType  );
    sURI      = QString( "myth://%1/%2" )
                   .arg( m_mapBackendIp  [ sHostName ] )
                   .arg( sBaseName );

    pRes = pItem->AddResource( sProtocol, sURI );

    pRes->AddAttribute( "duration"  , sDur      );
    pRes->AddAttribute( "size"      , longLongToString( nFileSize) );

    // ----------------------------------------------------------------------
    // Add Thumbnail Resource
    // ----------------------------------------------------------------------

    sURI = QString( "%1getPreviewImage%2").arg( sURIBase   )
                                          .arg( sURIParams ); 

    pItem->AddResource( "http-get:*:image/png:*" , sURI );
}

